import 'dart:io';

import 'package:dart_bones/dart_bones.dart';
import 'package:filesystem_picker/filesystem_picker.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:image/image.dart' as _image;
import 'package:sprintf/sprintf.dart';

import '../helper/file_supplier.dart';
import '../helper/validators.dart' as validators;
import 'application_data.dart';

class PageScalator extends StatefulWidget {
  final ApplicationData applicationData;
  PageScalator(this.applicationData, {Key key}) : super(key: key);

  @override
  PageScalatorState createState() {
    final rc = PageScalatorState(applicationData);
    return rc;
  }
}

class PageScalatorState extends State<PageScalator> {
  static RegExp regExpJPeg = RegExp(r'\.jpe?g$', caseSensitive: false);

  final ApplicationData applicationData;

  final GlobalKey<FormState> _formKey =
      GlobalKey<FormState>(debugLabel: 'collection');
  final sourceDirectoryController = TextEditingController();
  final targetDirectoryController = TextEditingController();

  PageScalatorState(this.applicationData);

  @override
  Widget build(BuildContext context) {
    final padding = 16.0;
    final applicationData = ApplicationData();
    sourceDirectoryController.text = applicationData.sourceDirectory.path;
    targetDirectoryController.text = applicationData.targetDirectory.path;
    final fileList = applicationData.fileList.map((e) => Text(e)).toList();
    final rc = Scaffold(
        appBar: applicationData.appBarBuilder('Bilder skalieren'),
        drawer: applicationData.drawerBuilder(context),
        body: Form(
          key: _formKey,
          child: Card(
              margin:
                  EdgeInsets.symmetric(vertical: padding, horizontal: padding),
              child: Padding(
                  padding: EdgeInsets.symmetric(
                      vertical: padding, horizontal: padding),
                  child: Column(
                    children: <Widget>[
                      Row(
                        children: [
                          Expanded(
                            child: TextFormField(
                              validator: validators.checkDirectory,
                              decoration: InputDecoration(
                                  labelText: 'Quellverzeichnis'),
                              onSaved: (input) =>
                                  applicationData.sourceDirectory.path = input,
                              controller: sourceDirectoryController,
                            ),
                          ),
                          Container(
                              width: 100,
                              child: RaisedButton(
                                onPressed: () => selectDirectory(
                                    context, applicationData.sourceDirectory),
                                child: Text('Auswahl'),
                              )),
                        ],
                      ),
                      Row(
                        children: [
                          Expanded(
                            child: TextFormField(
                              validator: validators.checkDirectory,
                              decoration:
                                  InputDecoration(labelText: 'Zielverzeichnis'),
                              onSaved: (input) =>
                                  applicationData.targetDirectory.path = input,
                              controller: targetDirectoryController,
                            ),
                          ),
                          Container(
                            width: 100,
                            child: RaisedButton(
                              onPressed: () => selectDirectory(
                                  context, applicationData.targetDirectory),
                              child: Text('Auswahl'),
                            ),
                          )
                        ],
                      ),
                      TextFormField(
                        validator: (input) =>
                            RegExp(r'^\w+(,\w+)*$').firstMatch(input) == null
                                ? "Keine Liste von Dateiendungen: $input"
                                : null,
                        initialValue: 'jpg,png',
                        decoration: InputDecoration(
                            labelText: 'Dateiendungen (mit Komma getrennt)'),
                        onSaved: (input) => applicationData.extensions =
                            '(' + input.replaceAll(',', '|') + r')$',
                      ),
                      TextFormField(
                        initialValue: '800x600',
                        validator: validators.checkDimensions,
                        decoration: InputDecoration(
                            labelText:
                                'Hochformat: Maximalbreite x Maximalhöhe'),
                        onSaved: (input) => applicationData.portrait =
                            ImageDimensions.fromString(input),
                      ),
                      TextFormField(
                        initialValue: '800x600',
                        validator: validators.checkDimensions,
                        decoration: InputDecoration(
                            labelText:
                                'Querformat: Maximalbreite x Maximalhöhe'),
                        onSaved: (input) => applicationData.landscape =
                            ImageDimensions.fromString(input),
                      ),
                      TextFormField(
                          initialValue: '600',
                          validator: validators.checkNat,
                          decoration: InputDecoration(
                              labelText: 'Quadratformat: Maximalbreite'),
                          onSaved: (input) => applicationData.square =
                              ImageDimensions.fromString('$input x $input')),
                      SizedBox(
                        height: padding,
                      ),
                      RaisedButton(
                        onPressed: () => run(context),
                        child: Text('Los gehts'),
                      ),
                      SizedBox(
                        height: padding,
                      ),
                      TextField(
                        controller: applicationData.messageController,
                        readOnly: true,
                      ),
                      SizedBox(
                        height: padding,
                      ),
                      Text('Meldungen'),
                      Expanded(
                          child: Card(
                              margin: EdgeInsets.symmetric(
                                  vertical: padding, horizontal: padding),
                              child: Padding(
                                  padding: EdgeInsets.symmetric(
                                      vertical: padding, horizontal: padding),
                                  child: ListView(
                                    children: fileList,
                                  ))))
                    ],
                  ))),
        ));
    return rc;
  }

  @override
  void dispose() {
    // Clean up the controller when the widget is removed from the
    // widget tree.
    sourceDirectoryController.dispose();
    targetDirectoryController.dispose();
    super.dispose();
  }

  void redraw() {
    setState(() => 1);
  }

  void run(context) async {
    if (_formKey.currentState.validate()) {
      _formKey.currentState.save();
      final applicationData = ApplicationData();
      applicationData.messageController.text = 'Ich laufe...';
      final supplier = FileSupplier(
          filePatterns: [applicationData.sourceDirectory.path],
          fileOptions: FileOptions(
              processBinaries: true,
              recursive: true,
              yieldDirectory: false,
              yieldLinkToDirectory: false,
              included:
                  RegExp(applicationData.extensions, caseSensitive: false)),
          verboseLevel: 4);
      applicationData.fileList.clear();
      try {
        for (var filename in supplier.next()) {
          final entry = supplier.currentEntity as File;
          final size = FileSync.humanSize(entry.lengthSync());
          //applicationData.fileList.add("$filename $size");
          //redraw();
          try {
            scaleOne(entry, supplier.relPath());
          } on Exception catch (exc) {
            applicationData.logger.error(exc.toString());
          }
        }
      } on Exception catch(exc){
        applicationData.logger.error('$exc');
      }
      log(sprintf('Bearbeitet: %d Datei(en) (mit %s) in %d Verzeichnis(sen) Ersparnisfaktor: %.0f%% Nur kopiert: %d Datei(en)', [
        applicationData.processedFiles,
        FileSync.humanSize(applicationData.processedBytes),
        supplier.processedDirs,
        applicationData.convertedBytes * 100.0/applicationData.processedBytes,
        applicationData.copiedFiles
      ]));
    }
  }

  void scaleOne(File source, String relPath) {
    // Read an image from file (webp in this case).
    // decodeImage will identify the format of the image and use the appropriate
    // decoder.
    _image.Image image = _image.decodeImage(source.readAsBytesSync());

    _image.Image newImage;
    ImageDimensions dimensions;
    int width, height;
    if (image == null) {
      applicationData.logger.error('unbekanntes Format in ${source.path}');
    } else {
      if ((width = image.width) == (height = image.height)) {
        dimensions = applicationData.square;
      } else if (width < height) {
        dimensions = applicationData.portrait;
      } else {
        dimensions = applicationData.landscape;
      }
      if (width > dimensions.maxWidth) {
        newImage = _image.copyResize(image, width: dimensions.maxWidth);
      } else if (height > dimensions.maxHeight) {
        newImage = _image.copyResize(image, height: dimensions.maxHeight);
      }
      String message;
        final target = FileSync.joinPaths(applicationData.targetDirectory.path,
            relPath, FileSync.nodeOf(source.path));
        FileSync.ensureDirectory(FileSync.parentOf(target));
      if (newImage == null) {
        if (width != null) {
          source.copy(target);
          applicationData.copiedFiles++;
          log(message = '${source.path} ${width}x$height');
        }
      } else {
        File file = File(target);
        if (regExpJPeg.firstMatch(target) != null) {
          file.writeAsBytesSync(
              _image.encodeJpg(newImage, quality: applicationData.quality));
        } else {
          file.writeAsBytesSync(_image.encodePng(newImage));
        }
        final size = file.lengthSync();
        final sourceSize = FileSync.humanSize(source.lengthSync());
        final targetSize = FileSync.humanSize(file.lengthSync());
        applicationData.convertedBytes += size;
        applicationData.processedFiles++;
        applicationData.processedBytes += source.lengthSync();

        log('${source.path} ${width}x$height $sourceSize -> $targetSize');
      }
    }
  }

  /// Logs a [message] to the logger and to the screen.
  void log(String message) {
    applicationData.logger.log(message);
    applicationData.messageController.text = message;
    if (applicationData.fileList.length > 20){
      applicationData.fileList.removeAt(0);
    }
    applicationData.fileList.add(message);
    redraw();
  }

  /// Selects a directory with an dialog.
  /// [dirInfo] defines the start directory.
  /// The selection was stored into [dirInfo].
  void selectDirectory(context, DirInfo dirInfo) async {
    dirInfo.path = await FilesystemPicker.open(
      title: '${dirInfo.label} auswählen',
      context: context,
      rootDirectory: Directory(dirInfo.path),
      fsType: FilesystemType.folder,
      pickText: 'Den aktuellen Ordner auswählen',
      //folderIconColor: Colors.teal,
    );
    redraw();
  }
}
